# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class Template2LevelMVNormalRSCC(Template):
    'Fits 2-level multivariate Normal response models, allowing for random slopes, missing data, and a choice of uniform or Wishart priors. Creates custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'Normal', 'Multivariate response', 'Random slopes']
    engines = ['CustomC']

    inputs = '''
D = Text(value = 'MVNormal')
y = DataMatrix('Responses: ')
L2ID = IDVector('Level 2 ID: ')
for i in range(0, len(y)):
    context['xf'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)
    context['xr'+str(i+1)] = DataMatrix('Explanatory variables random for response ' + y[i]+ '?: ', allow_cat = True)

storeresid = Boolean('Store residuals?')
randu = Boolean('Randomise starting value for level 2 residuals?')

lenbeta = 0
lenu = 0
for i in range(0, len(y)):
    lenbeta += len(context['xf'+str(i+1)])
    context['miss'+y[i]] = ParamVector(monitor=False)
    lenu += len(context['xr'+str(i+1)])
    for x in range(0, len(context['xr'+str(i+1)])):
        if storeresid:
            context['u' + str(x) + '_' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False)
        else:
            context['u' + str(x) + '_' + str(i)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)

n = len(y)
if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    tau_u = ParamScalar()
    sigma_u = ParamScalar(modelled = False)
else:
    omega_e = ParamMatrix(modelled = False)
    omega_e.size = n
    omega_u = ParamMatrix(modelled = False)
    omega_u.size = lenu
    d_e = ParamMatrix()
    d_e.size = n
    d_u = ParamMatrix()
    d_u.size = lenu
    priors = Text('Priors: ', ['Uniform', 'Wishart'])
    if priors == 'Wishart':
        R = List('R matrix: ')
        v = Integer('Degrees of Freedom:')
    priors2 = Text('Level 2 Priors: ', ['Uniform', 'Wishart'])
    if priors2 == 'Wishart':
        R2 = List('R matrix: ')
        v2 = Integer('Degrees of Freedom:')


beta = ParamVector()
beta.ncols = lenbeta

deviance = ParamScalar(modelled = False)
'''

    model = '''
<% lenbeta = 0 %>
<% lenu = 0 %>
<% n = len(context['y']) %>
model {
    for (i in 1:length(${y[0]})) {
    % if n == 1:
            miss${y[0]}[i] ~ dnorm(mu0[i], tau)
            mu0[i] <- ${mmult2(context['xr1'], 'beta', 'i', lenbeta)} + u0[${L2ID}[i]]
<% lenbeta += len(context['xf1']) %> \\
    % else:
            dummy[i] ~ dnormal${n}a( \\
        % for i in range(0,n):
miss${y[i]}[i], \\
        % endfor
        % for i in range(0,n):
mu${i}[i], \\
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_e[${i}], \\
        % endfor
d_e[${(((n+1)*n)/2) - 1}])
        % for i in range(0, len(y)):
            mu${i}[i] <- ${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)} \\

            % for j in range(0, len(context['xr'+str(i+1)])):
            + u${j}_${i}[${L2ID}[i]] * ${context['xr'+str(i+1)][j]}[i]
            % endfor
<% lenbeta += len(context['xf' + str(i+1)]) %> \\
<% lenu += len(context['xr' + str(i+1)]) %> \\
        % endfor
        % for i in range(0,n):
        miss${y[i]}[i] ~ dflat()
        % endfor
    % endif
    }
      
    for (j in 1:length(u0_0)) {
        %if len(y) == 1 :
           u0[j] ~ dnorm(0, tau_u)
    % else:
        dummy_u[j] ~ dnormal${lenu}a( \\
% for i in range(0, len(y)):
 % for j in range(0, len(context['xr'+str(i+1)])):
u${j}_${i}[j], \\
        % endfor
        % endfor
        % for i in range(0,lenu):
0, \\
        % endfor
        % for i in range(0,(((lenu+1)*lenu)/2) - 1):
d_u[${i}], \\
        % endfor
d_u[${(((lenu+1)*lenu)/2) - 1}])

        dum_u[j] ~ ddummy(dummy_u[j])
% for i in range(0, len(y)):
 % for j in range(0, len(context['xr'+str(i+1)])):
        u${j}_${i}[j] ~ dflat()
        % endfor
        % endfor

    % endif

    }

    # Priors
    
    % if n == 1:
    tau ~ dgamma(0.001, 0.001)
    sigma <- 1 / sqrt(tau)
    tau_u ~ dgamma(0.001, 0.001)
    sigma_u <- 1 / sqrt(tau_u)
    % endif

    % for i in range(0, lenbeta):
    beta_${i} ~ dflat()
    % endfor   
}       
''' 

    latex = r'''
<% lenbeta = 0 %>
<% n = len(y) %>
\begin{aligned} 
\left(
\begin{array}{l}
% for r in range(0, len(y)):
\mbox{${y[r]}}_i 
% if r != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right) & \sim \mbox{N}
\left[ \left(
\begin{array}{l}
% for r in range(0, len(y)):
\mu_{${r}i}
% if r != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right), \Omega_e \right] \\ 
% for r in range(0, len(y)):
\mu_{${r}i} & = ${mmulttex2(context['xf' + str(r+1)], r'\beta', 'i', lenbeta)}
% for j in range(0, len(context['xr' + str(r + 1)])):
+ u_{${r},${j},${L2ID}[i]}^{(2)} \mbox{${context['xr' + str(r + 1)][j]}}_{i}
% endfor
\\ 
<% lenbeta += len(context['xf' + str(r+1)]) %>
% endfor

\left(
\begin{array}{l}
% for r in range(0, len(y)):
% for j in range(0, len(context['xr' + str(r + 1)])):
u^{(2)}_{${r}, ${j},${L2ID}(i)} 
% if j != len(context['xr' + str(r + 1)]) - 1 or r != len(y) - 1:
\\ 
% endif
% endfor
% endfor
\end{array}
\right) & \sim \mbox{N}

\left[ \left(
\begin{array}{l}
% for r in range(0, len(y)):
% for i in range(0, len(context['xr' + str(r + 1)])):
0
% if i != len(context['xr' + str(r + 1)]) -1 or r != len(y) - 1:
\\ 
% endif
% endfor
% endfor
\end{array}
\right), \Omega^{(2)}_{u} \right] \\ 


<% lenu = 0 %>
% for i in range(0, len(y)):
<% lenu += len(context['xr' + str(i+1)]) %>
% endfor

% if priors2 == 'Wishart':
\Omega^{(2)}_u & \sim \mbox{inverse Wishart}_{${lenu}}[${v2}*R_{u2},${v2}],R_{u2}=
\left[\begin{array}{
% for i in range(0, lenu):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, lenu):
% for j in range(0, lenu):
% if j <= i:
${R2.name[lenR]}
<% lenR += 1 %>
% endif
% if j != lenu - 1:
&
% else:
% if i != lenu - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if priors2 == 'Uniform':
\Omega^{(2)}_u & \propto 1
% endif
\\ 


% if priors == 'Wishart':
\Omega_e & \sim \mbox{inverse Wishart}_{${len(y)}}[${v}*R_e,${v}],R_e=
\left[\begin{array}{
% for i in range(0, len(y)):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, len(y)):
% for j in range(0, len(y)):
% if j <= i:
${R.name[lenR]}
<% lenR += 1 %>
% endif
% if j != len(y) - 1:
&
% else:
% if i != len(y) - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if priors == 'Uniform':
\Omega_e & \propto 1
% endif
\\ 

% for i in range(0, lenbeta):
\beta_${i} & \propto 1 \\ 
% endfor   

\end{aligned} 
'''

    customccode = '''
        <%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out
%>
    {
<% n = len(y) %>\\
        // Note currently using a uniform prior for variance matrix
        SymMatrix sbe(${n});
        int vwe = length(${y[0]});
        for(int i = 0; i < length(${y[0]}); i++) {
<% lenbeta = 0 %>\\
% for i in range(0, n):
            double e${i} = double(miss${y[i]}[i]) - (${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)}) \\
% for j in range(0, len(context['xr'+str(i+1)])):
 - double(u${j}_${i}[int(${L2ID}[i])]) * double(${context['xr' + str(i+1)][j]}[i])\\
%endfor
;\\
<% lenbeta += len(context['xf' + str(i + 1)]) %>
% endfor

% for i in range(0, n):
% for j in range(i, n):
            sbe(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }
% if priors == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbe(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(i, n):
        sbe(${i}, ${j}) += ${str(Rmat[i][j] * float(v))};
% endfor
% endfor
% endif

        matrix_sym_invinplace(sbe);

% if priors == 'Uniform':
        vwe -= ${n + 1};
% endif
% if priors == 'Wishart':
        vwe += ${v};
% endif

        bool fail = false;
        mat_d_e = dwishart(vwe, sbe, fail);
        mat_omega_e = matrix_sym_inverse(mat_d_e, fail);


        std::vector<double*> tmp_u;
% for i in range(0, len(y)):
% for j in range(0, len(context['xr'+str(i+1)])):
        tmp_u.push_back(u${j}_${i});
% endfor
% endfor
        RectMatrix mat_u(tmp_u, length(u0_0));

        SymMatrix sbu = mat_u.T() * mat_u;

% if priors2 == 'Uniform':
        if (runstate == 0) {
% for i in range(0, len(y)):
            sbu(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors2 == 'Wishart':
<%
import numpy
Rmat = numpy.empty([len(y), len(y)])
count = 0 
for i in range(0, len(y)):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R2[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, len(y)):
% for j in range(i, len(y)):
        sbu(${i}, ${j}) += ${str(Rmat[i, j] * float(v2))};
% endfor
% endfor
% endif
% if priors2 == 'Uniform':
        int vwu = length(u0_0) - ${len(y) + 1};
% endif
% if priors2 == 'Wishart':
        int vwu = length(u0_0) + ${v2};
% endif

        matrix_sym_invinplace(sbu);
        mat_d_u = dwishart(vwu, sbu, fail); 
        mat_omega_u = matrix_sym_inverse(mat_d_u, fail);



                // Step for beta
                std::vector<int> betaind(${lenbeta});
<% betacount = 0 %>
% for i in range(0, n):
% for j in range(0, len(context['xf' +str(i+1)])):
                betaind[${betacount}] = ${i};
<% betacount += 1 %>
% endfor
% endfor
                std::vector<double*> tmp_x;
% for i in range(0, n):
% for j in range(0, len(context['xf' + str(i+1)])):
                tmp_x.push_back((double*)${context['xf' + str(i+1)][j]});
% endfor
% endfor
                RectMatrix mat_x(tmp_x, length(${y[0]}));


                std::vector<double*> tmp_xr;
% for i in range(0, n):
% for j in range(0, len(context['xr' + str(i+1)])):
                tmp_xr.push_back((double*)${context['xr' + str(i+1)][j]});
% endfor
% endfor
                RectMatrix mat_xr(tmp_xr, length(${y[0]}));

                std::vector<double*> tmp_ymiss;
% for i in range(0, n):
                tmp_ymiss.push_back(miss${y[i]});
% endfor
                RectMatrix mat_ymiss(tmp_ymiss, length(${y[0]}));

                SymMatrix db(${lenbeta});
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db(j1, j2) += mat_x(i, j1) * mat_d_e(betaind[j1], betaind[j2]) * mat_x(i, j2);
                        }
                    }
                }

                RectMatrix xtdy(${lenbeta}, 1);
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                            
<% count = 0 %>
% for j2 in range(0, n):
                            double etemp${j2} = mat_ymiss(i, ${j2});
% for k in range(0, len(context['xr'+str(j2+1)])):
                            etemp${j2} -= mat_xr${j2 + 1}(i, ${k}) * mat_u(${L2ID}[i], ${count});
<% count += 1 %>
% endfor
                            xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], ${j2}) * etemp${j2}; 

% endfor
                    }
                }

                fail = false;
                // int err = matrix_sym_invinplace(db);
                // RectMatrix betahat = db * xtdy;
                // mat_beta = dmultnormal(betahat, db, fail);

                RectMatrix betahat = matrix_solve(db, xtdy);
                mat_beta = dmultnormalinv(betahat, db, fail);

                std::vector<int> betamap(${n});
                int pos = 0;
% for i in range(0, n):
                betamap[${i}] = pos;
                pos += mat_xf${i + 1}.cols();

% endfor


                std::vector<int> umap(${n});
                int upos = 0;
% for i in range(0, n):
                umap[${i}] = upos;
                upos += mat_xr${i + 1}.cols();
% endfor

<% lenu = 0 %>\\
% for i in range(0, len(y)):
<% lenu += len(context['xr' + str(i+1)]) %>\\
% endfor
                std::vector<int> uind(${lenu});
<% ucount = 0 %>
% for i in range(0, n):
% for j in range(0, len(context['xr' +str(i+1)])):
                uind[${ucount}] = ${i};
<% ucount += 1 %>
% endfor
% endfor

                for (int i = 0; i < mat_u.rows(); i++) {
                    SymMatrix duj(mat_u.cols());
                    for (int j = 0; j < mat_u.cols(); j++) {
                        for (int k = 0; k <= j; k++) {
                            duj(j, k) = mat_d_u(j, k);
                            for (int id_m = 0; id_m < idmap_${L2ID}[i].size(); id_m++) {
                                int m = idmap_${L2ID}[i][id_m];
                                duj(j, k) += mat_xr(m, j) * mat_d_e(uind[j], uind[k]) * mat_xr(m, k);
                            }
                        }
                    }

                    RectMatrix tmpuhat(mat_u.cols(), 1);
                    for (int id_k = 0; id_k < idmap_${L2ID}[i].size(); id_k++) {
                        int k = idmap_${L2ID}[i][id_k];
                        RectMatrix ymxb(${n}, 1);
% for j in range(0, n):
                        ymxb(${j}, 0) = mat_ymiss(k, ${j});
                        for (int m = 0; m < mat_xf${j + 1}.cols(); m++) {
                            ymxb(${j}, 0) -= mat_xf${j + 1}(k, m) * mat_beta(betamap[${j}] + m, 0);
                        }
% endfor
                        
                        for (int j = 0; j < mat_u.cols(); j++) {
                            for (int j1 = 0; j1 < ${n}; j1++) {
                                tmpuhat(j, 0) += mat_xr(k, j) * mat_d_e(uind[j], j1) * ymxb(j1, 0);
                            }
                        }
                    }


                    // matrix_sym_invinplace(duj);
                    // RectMatrix uhat = matrix_mult(duj, tmpuhat);
                    // RectMatrix newuj = dmultnormal(uhat, duj, fail);

                    RectMatrix uhat = matrix_solve(duj, tmpuhat);
                    RectMatrix newuj = dmultnormalinv(uhat, duj, fail);

                    for (int j = 0; j < mat_u.cols(); j++) {
% if randu:
                        if (runstate != 0) {
                            mat_u(i, j) = newuj(j, 0);
                        } else {
                            mat_u(i, j) = dnorm(0, 1);
                        }
% else:
                        mat_u(i, j) = newuj(j, 0);
% endif
                    }
                }



% for i in range(0, n):
                for (int j = 0; j < mat_ymiss.rows(); j++) {
                    if (mat_y(j, ${i}) <= -9.999e29) {

                        double tmp = 0.0;
                        for (int k = 0; k < mat_xf${i + 1}.cols(); k++) {
                            tmp += mat_beta(betamap[${i}] + k, 0) * mat_xf${i + 1}(j, k);
                        }
                        for (int k = 0; k < mat_xr${i + 1}.cols(); k++) {
                            tmp += mat_u(${L2ID}[j], umap[${i}] + k) * mat_xr${i + 1}(j, k);
                        }


                        for (int k = 0; k < mat_ymiss.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e(${i}, k) * mat_ymiss(j, k) * (1.0 / mat_d_e(${i}, ${i}));
                            }
                        }
% for m in range(0, n):
% if i != m:
                        for (int k = 0; k < mat_xf${m + 1}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1 / mat_d_e(${i}, ${i})) * mat_beta(betamap[${m}] + k, 0) * mat_xf${m + 1}(j, k);
                        }

                        for (int k = 0; k < mat_xr${m + 1}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1 / mat_d_e(${i}, ${i})) * mat_u(${L2ID}[j], umap[${m}] + k) * mat_xr${m + 1}(j, k);
                        }

% endif
% endfor
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    }
                }
% endfor

    }
'''

    preparedata = '''
mydata = data['datafile']

for i in range(0, len(y)):
    context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
'''
